/*
 all-ic.c

 Parallel version of simple irregular loop (see `seq.c' file for details).
 All arrays are in-core (they fit into the memory so they are not stored on
 disk).
 For simplicity of the code all arrays are assumed to be evenly divisible among
 all nodes.
 */

#include <stdio.h>
#include <stdlib.h>

#include <lip.h>
#include <lipimpl.h>
#include <debug.h>
#include "generate.h"

/* size of data array can be evenly divided by number 1...10 (which are allowed
 * total numbers of nodes) */
#define N (2 * 3 * 2 * 5 * 7 * 2 * 9)
//#define N (2 * 3 * 2)
#define L (2 * N)
/* ceiling division from HPF 2.0 */
#define CD(j,k) (((j)-1)/(k)+1)


int
main(int argc, char **argv)
{
  double *x, *y;
  int *perm, *perm_l;
  int n = N,   l = L;
  int n_l,     l_l;    /* numbers of elements on single node (local elements) */
  int n_start, l_start; /* first index on this node */

  int i, s, p,l_loc;
  double sum_y_l, sum_y, sum_x_l, sum_x, z;

  int rank, size; /* LIP's environment parameters */

  LIP_Datamap datamap;
  LIP_Schedule schedule;

  MPI_Init( &argc, &argv );
  LIP_Setup( NULL, 0 ); /* all processes go to LIP */

  LIP_Rank( &rank );
  LIP_Size( &size );

  l_l = CD( l, size ); /* number of `x' elements on single node */
  n_l = CD( n, size ); /* number of `y' & `perm' elements on single node */
  l_loc=l_l;
  l_start = l_l * rank;
  n_start = n_l * rank;
  if(rank==size-1){
    l_l=l-l_l*rank;
    n_l=n-n_l*rank;
  }

  /*** Allocate  memory ***/

  x    = malloc( (l_l + n_l) * (sizeof *x) ); /* additional space for non-local
                                            data (which may be of size `n_l') */
  y    = malloc( n_l * (sizeof *y) );

  perm   = malloc( n_l * (sizeof *perm) );
  perm_l = malloc( n_l * (sizeof *perm) );

  /*** Fill data  array ***/
  generate_data(x, perm, l, l_start, l_start+l_l, n_start, n_start+n_l);
  /*** Inspector  phase ***/

  /* create data map which describes how data array `x' is partitioned */
  LIP_Datamap_create( LIP_DATAMAP_BLOCK, l, &l_loc, LIP_INDEXTYPE_INT, 1, 0,
    &datamap );

  LIP_Schedule_create( l_l, &schedule );

  LIP_Localize( datamap, perm, LIP_INDEXTYPE_INT, perm_l, LIP_INDEXTYPE_INT,
    n_l, 0, &schedule, LIP_INDEXINFO_NULL );

  LIP_Datamap_free( &datamap, 0);

  /* schedule has to be committed before its use in communication functions */
  LIP_Schedule_commit( &schedule );

  /*** Executor   phase ***/

  /* gather non-local data from other nodes */
  LIP_Gather( x, x + l_l, MPI_DOUBLE, schedule );

  /* irregular loops */

  for (i = 0 ; i < n_l ; i++)
    y[i] = -x[ perm_l[i] ];

  /* clear the ghost area */
  for (i = 0 ; i < n_l ; i++)
    x[ i + l_l ] = 0.0;

  for (i = 0 ; i < n_l ; i++)
     x[ perm_l[i] ] += y[i];


  /* scatter non-local data to other nodes */
  LIP_Scatter( x + l_l, x, MPI_DOUBLE, schedule, LIP_OP_SUM );

  LIP_Schedule_free( &schedule );

  if (0 == rank && size > 1)
    printf( "%d computing nodes. "
      "Each has %d `y' elements and %d `x' elements.\n", size, n_l, l_l );

  for (sum_y_l = 0.0, i = 0 ; i < n_l ; i++)
    sum_y_l += y[i];

  MPI_Reduce( &sum_y_l, &sum_y, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `y' is %g\n", n, sum_y );

  for (sum_x_l = 0.0, i = 0 ; i < l_l ; i++)
    sum_x_l += x[i];

  MPI_Reduce( &sum_x_l, &sum_x, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `x' is %g\n", l, sum_x );

  LIP_Exit();
  MPI_Finalize();

  free( perm_l );
  free( perm );
  free( y );
  free( x );

  return 0;
} /* main */
