/*
 data-ooc.c

 Parallel version of simple irregular loop (see `seq.c' file for details).
 Index arrays are in-core (they fit into the memory so they are not stored on
 disk).
 Data arrays are out of core so they are to be read from disk using MPI IO
 routines
 For simplicity of the code all arrays are assumed to be evenly divisible among
 all nodes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <debug.h>
#include "generate.h"
#include <lip.h>

/* size of data array can be evenly divided by number 1...10 (which are allowed
 * total numbers of nodes) */
#define N (2 * 3 * 2 * 5 * 7 * 2 * 9)
/*#define N (2 * 3 * 2)*/
#define L (2 * N)

/* ceiling division from HPF 2.0 */
#define CD(j,k) (((j)-1)/(k)+1)


int
main(int argc, char **argv)
{
  double *x, *y, *x_l;
  int *perm, *perm_l, *tperm_l;
  int n = N,   l = L;
  int n_l,     l_l;    /* numbers of elements on single node (local elements) */
  int n_start, l_start; /* first index on this node */

  int i, ret;
  double sum_y_l, sum_y, sum_x_l, sum_x;

  int rank, size; /* LIP's environment parameters */
  
  MPI_Info info; /* needed by MPI IO */
  MPI_File file;
  MPI_Offset offset;
  MPI_Status status;


  LIP_Datamap datamap, filetype, memtype;
  LIP_Schedule schedule;
  LIP_IOBufmap bufm;	

  MPI_Init( &argc, &argv );
  LIP_Setup( NULL, 0 ); /* all processes go to LIP */
  MPI_Info_create(&info);

  LIP_Rank( &rank );
  LIP_Size( &size );

  l_l = CD( l, size ); /* number of `x' elements on single node */
  n_l = CD( n, size ); /* number of `y' & `perm' elements on single node */

  l_start = l_l * rank;
  n_start = n_l * rank;

  /*** Allocate  memory ***/

  x    = malloc( (l_l + n_l) * (sizeof *x) ); /* additional space for non-local
                                            data (which may be of size `n_l') */
  y    = malloc( n_l * (sizeof *y) );

  x_l    = malloc( (n_l + n_l) * (sizeof *x) ); /* additional space for non-local
                                            data (which may be of size `n_l') */

  perm   = malloc( n_l * (sizeof *perm) );
  perm_l = malloc( n_l * (sizeof *perm) );
  tperm_l = malloc( n_l * (sizeof *perm) );
  /*** Fill data  array ***/

  generate_data(x, perm, l, l_start, l_start+l_l, n_start, n_start+n_l);

/*    x[i] = l_start+i;*/
  
  
  /*** Fill data file ***/
  MPI_File_open(MPI_COMM_WORLD, "data.file", MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &file);
  MPI_File_preallocate(file,L*sizeof(*x));
  MPI_File_set_view(file,0,MPI_DOUBLE, MPI_DOUBLE, "native", MPI_INFO_NULL);
  offset = rank * l_l; /* Each node writes its part of a file */
  MPI_File_write_at(file,offset,x,l_l,MPI_DOUBLE,&status);
/*  
  printf ("Rank %i - Array x:\n",rank);
  for (i = 0 ; i < l_l ; i++) 
    printf ("%2.2f ",x[i]);
  printf ("\n\n");
*/  

/*  
  printf ("Rank %i - Index array perm:\n",rank);
  for (i = 0 ; i < n_l ; i++) 
    printf ("%i ",perm[i]);
  printf ("\n\n");
*/
  /*** Inspector  phase ***/

  /* create data map which describes how data array `x' is partitioned */
  LIP_Datamap_create( LIP_DATAMAP_BLOCK, l, &l_l, LIP_INDEXTYPE_INT, 1, 0,
    &datamap );
  /* create IO bufmap which maps local data buffer on the data array on disk */  
  LIP_IOBufmap_create(n_l, &bufm);

  LIP_Schedule_create( l_l, &schedule );

  LIP_Localize( datamap, perm, LIP_INDEXTYPE_INT, perm_l, LIP_INDEXTYPE_INT,
    n_l, 0, &schedule, LIP_INDEXINFO_NULL );
/*
  printf ("Rank %i - Local index array perm_l:\n",rank);
  for (i = 0 ; i < n_l ; i++) 
    printf ("%i ",perm_l[i]);
  printf ("\n\n");
  fflush(NULL);
*/
    
  LIP_OOC_localize(perm_l, LIP_INDEXTYPE_INT, tperm_l, LIP_INDEXTYPE_INT, 
  		n_l, 0, x_l, MPI_DOUBLE, &bufm, &schedule);
/*
  printf ("Rank %i - Transformed local index array tperm_l:\n", rank);
  for (i = 0 ; i < n_l ; i++) 
    printf ("%i ",tperm_l[i]);
  printf ("\n\n");
  fflush(NULL);
*/
  LIP_Datamap_free( &datamap, 0 );

  /* schedule has to be committed before its use in communication functions */
  LIP_Schedule_commit( &schedule );
  
  /* get datatypes mapping memory to disk */

    ret = LIP_IOBufmap_get_datatype(bufm,MPI_DOUBLE,&memtype,&filetype,LIP_DATA_NEW);
    if (1)
    {
      ret = MPI_File_set_view(file,l_l*rank*8,MPI_DOUBLE,filetype,"native",info);
      /* read from files */
      ret = MPI_File_read_at(file,0, x_l, 1, memtype, &status);
      printf ("Rank %i - aqq", rank);
    }  

     else
     {
       MPI_File_set_view(file,l_l*rank*8,MPI_DOUBLE, MPI_DOUBLE,"native",info);
     }
/*  
  printf ("Rank %i - Local array x_l read from disk:\n", rank);
  for (i = 0 ; i < l_l ; i++) 
    printf ("%2.2f ",x_l[i]);
  printf ("\n\n");
*/

  /*** Executor   phase ***/

  /* gather non-local data from other nodes */
  LIP_Gather( x_l, x_l + n_l, MPI_DOUBLE, schedule );
/*  
  printf ("Rank %i - Local array x_l with Ghost area:\n", rank);
  for (i = 0 ; i < l_l+n_l ; i++) 
    printf ("%2.2f ",x_l[i]);
  printf ("\n\n");
*/

  /* irregular loops */

  for (i = 0 ; i < n_l ; i++)
    y[i] = -x_l[ tperm_l[i] ];

  /* clear the ghost area */
  for (i = 0 ; i < n_l ; i++)
    x_l[ i + n_l ] = 0.0;
   
  for (i = 0 ; i < n_l ; i++)
     x_l[tperm_l[i] ] += y[i];
     
     

  /* scatter non-local data to other nodes */
  LIP_Scatter( x_l + n_l, x_l, MPI_DOUBLE, schedule, LIP_OP_SUM );

  LIP_Schedule_free( &schedule );

  /* modify data values on the disk*/
  LIP_IOBufmap_get_datatype(bufm,MPI_DOUBLE,&memtype,&filetype,LIP_DATA_ALL);
  MPI_File_set_view(file,l_l*rank*8,MPI_DOUBLE,filetype,"native",info);
  MPI_File_write_at(file,0, x_l, 1, memtype, &status);

  /* read modified x array */  
  MPI_File_set_view(file,0,MPI_DOUBLE, MPI_DOUBLE, "native", MPI_INFO_NULL);
  offset = rank * l_l; /* Each node writes its part of a file */
  MPI_File_read_at(file,offset,x,l_l,MPI_DOUBLE,&status);



  if (0 == rank && size > 1)
    printf( "%d computing nodes. "
      "Each has %d `y' elements and %d `x' elements.\n", size, n_l, l_l );

  for (sum_y_l = 0.0, i = 0 ; i < n_l ; i++)
    sum_y_l += y[i];

  MPI_Reduce( &sum_y_l, &sum_y, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `y' is %g\n", n, sum_y );

  for (sum_x_l = 0.0, i = 0 ; i < l_l ; i++)
    sum_x_l += x[i];

  MPI_Reduce( &sum_x_l, &sum_x, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `x' is %g\n", l, sum_x );

  MPI_File_close(&file); /* Close a file before calling MPI_Finalize! */

  LIP_Exit();
  MPI_Finalize();


  free( perm_l );
  free( perm );
  free( y );
  free( x_l );

  return 0;
} /* main */
