/*
 datamap.c
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <lipimpl.h>
#include <lip.h>
#include <debug.h>
#define p(x) printf("[%d]" __FILE__ "(%d)" #x "=%g\n",LIP_env.rank,__LINE__,(double)(x)),fflush(0)
#define pa(a,i) printf("[%d]" __FILE__ "(%d)" #a "[%d]=%g\n",LIP_env.rank,__LINE__,(i),(double)((a)[i])),fflush(0)

#define LOCAL_DIR "/var/tmp/"

/* ceiling division (CD) as defined by HPF 2.0; `k' is evaluated twice ! */
/* #define CEILDIV(j,k) (((j)+(k)-1)/(k)) */
/* ceiling division in C (nodes and arrays are numbered starting with 0) */
#define CEILDIV(j,k) (j)/(k)
/* ceiling remainder in C (nodes and arrays are numbered starting with 0) */
#define CEILREM(j,k) (j)%(k)

/*
 Function: LIP_Datamap_create
 Descrption:
 Arguments:
   type - specifies type of datamap. Possible values are: LIP_DATAMAP_BLOCK
   data - its contents depends on 'type' argument
     LIP_DATAMAP_BLOCK
       data_size >= 1; data[0] = size of global data array for which datamap
       will be created
   data_size - number of elements in array 'data'
 Side effects:
   For some types of distribution (irregular) the function will call MPI
   collective communication routines.
 */
int
LIP_Datamap_create(int type, int global,void* desc,
  LIP_Indextype desctype,int count, 
  int flags, LIP_Datamap *datamap)
{
  LIP_IDatamap *imap;
  int *idesc;
  int m; /* `m' parameter for BLOCK(m) and CYCLIC(m) HPF 2.0 distributions */
  int left_over;
  char* pname="LIP_Datamap_create";
  /* FIXME: check values of parameters for HPF conformity */

  *datamap = LIP_DATAMAP_NULL;

  if (global <= 0){
    LIP_Error_ret(pname,"number of disc elements <=0",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
  /* if there is only one node */
  if (1 == LIP_env.size)
  {
    /* FIXME: may there should be special datamap for single node */
    return LIP_SUCCESS;
  }

  /* FIXME: general indexing */
  if (desctype != LIP_INDEXTYPE_INT){
    LIP_Error_ret(pname,"indextype",LIP_NOT_SUPPORTED);
    return LIP_NOT_SUPPORTED;
  }    
  idesc = desc;
  /* for HPF distibutions: BLOCK(m), CYCLIC(m) `m' is taken from the first
   * value of `desc' array */
  m = idesc[0];

  imap = LIP_Datamap_new( datamap );
 
  if(NULL==imap){
 
   return LIP_OUT_OF_MEMORY;
 
  }
  
  imap->type = type;
  
  imap->flag = flags;
  
  switch (type)
  {
    /* this creates HPF BLOCK(m) distibution (each processor is assigned `m'
     * elements, some processor may have less); for simple BLOCK distribution
     * (value of 'm' is computed according to size of global array) use
     * predefined LIP_DATAMAP_BLOCK datamap */
    case LIP_DATAMAP_BLOCK:

      if (m > global)
      {
        LIP_Datamap_free( datamap, 0 );

        *datamap = LIP_DATAMAP_NULL;

	LIP_Error_ret(pname,"local size greater than global size",LIP_INVALID_ARGUMENT);

        return LIP_INVALID_ARGUMENT; /* impossible */
      }
      else if (m * LIP_env.size < global)
      {
        LIP_Datamap_free( datamap, 0 );
    
        *datamap = LIP_DATAMAP_NULL;

	LIP_Error_ret(pname,"local sizes and global size not HPF 2.0 conformant",LIP_INVALID_ARGUMENT);

        return LIP_INVALID_ARGUMENT; /* not HPF 2.0 conformant */
      }

      /* global size of data array */
      imap->global_size = global;

      /* number of indices for single node */
      imap->local_size = m;

      /* compute number of indices for this node */

      left_over = global - LIP_env.rank * m; /* how many elements left */

      if (left_over >= m)
        imap->node_size = m;
      else
      {
        if (left_over < 0)
          imap->node_size = 0; /* no more data left */
	else
          imap->node_size = left_over; /* less than `m' data elements left */
      }
      
      imap->ttable_ic_size=0;
      
      imap->ttable_ic_real_size=0;
      
      imap->ttable_ooc_size=0;
      
      imap->procs=NULL;
      
      imap->local_adr=NULL;
      
      imap->procs_file=NULL;
      
      imap->local_adr_file=NULL;

      imap->procs_fname=NULL;
      
      imap->local_adr_fname=NULL;
    
      imap->tt_io_bmap=LIP_IOBUFMAP_NULL;

            
      break; /* case LIP_DTYPE_BLOCK: */
    
    case LIP_DATAMAP_INDIRECT :
      if (m > global)
      {
        LIP_Datamap_free( datamap, 0 );

        *datamap = LIP_DATAMAP_NULL;

	LIP_Error_ret(pname,"local size greater than global size",LIP_INVALID_ARGUMENT);

        return LIP_INVALID_ARGUMENT; /* impossible */

      }
      else if (m * LIP_env.size < global)
      {

        LIP_Datamap_free( datamap, 0);

        *datamap = LIP_DATAMAP_NULL;

	LIP_Error_ret(pname,"local sizes and global size not HPF 2.0 conformant",LIP_INVALID_ARGUMENT);

        return LIP_INVALID_ARGUMENT; /* not HPF 2.0 conformant */

       }

      /* global size of data array */
      imap->global_size = global;

      /* number of indices for single node */
      imap->local_size = m;

      /* compute number of indices for this node */

      left_over = global - LIP_env.rank * m; /* how many elements left */

      if (left_over >= m)
        imap->node_size = m;
      else
      {
        if (left_over < 0)
          imap->node_size = 0; /* no more data left */
	else
          imap->node_size = left_over; /* less than `m' data elements left */
      }

      imap->ttable_ic_size=0;

      imap->ttable_ic_real_size=0;

      imap->ttable_ooc_size=0;

      imap->procs=NULL;

      imap->local_adr=NULL;

      imap->procs_file=NULL;

      imap->local_adr_file=NULL;
      
      imap->procs_fname=NULL;
      
      imap->local_adr_fname=NULL;


      imap->tt_io_bmap=LIP_IOBUFMAP_NULL;
      
      /**** here we deal with ooc datamap ****/
      if(flags & LIP_OOC)
      {
           
          if(LIP_env.rank==0)
          {
                 imap->procs_fname = tempnam(".", "lip");
                 imap->local_adr_fname = tempnam(".", "lip");          
          }
          else/*rank !=0*/
          {
               imap->procs_fname = (char *) malloc(L_tmpnam);
               imap->local_adr_fname = (char *) malloc(L_tmpnam);
          }
          MPI_Bcast(imap->procs_fname, L_tmpnam, MPI_CHAR, 0, LIP_env.comm);
          MPI_Bcast(imap->local_adr_fname, L_tmpnam, MPI_CHAR, 0, LIP_env.comm);
      
            MPI_File_open(MPI_COMM_WORLD, imap->procs_fname, MPI_MODE_RDWR|
                          MPI_MODE_CREATE, MPI_INFO_NULL, &imap->procs_file);
            MPI_File_open(MPI_COMM_WORLD, imap->local_adr_fname,MPI_MODE_RDWR |
                          MPI_MODE_CREATE, MPI_INFO_NULL, &imap->local_adr_file);

      }

      break;
  }

  return LIP_SUCCESS;
}


int LIP_Datamap_ic_ttable_detach(LIP_Datamap datamap,
 int** procs,int** local_adr,int* ttable_size){
  LIP_IDatamap *imap;
  
  int *iprocs;
  int *ilocal_adr;
  
  char* pname="LIP_Datamap_ic_ttable_dettach";  
  
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
  
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
  
   return LIP_INVALID_ARGUMENT;
  
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
  
   LIP_Error_ret(pname,"you can dettach table only from INDIRECT datamap",LIP_INVALID_ARGUMENT);
  
   return LIP_INVALID_ARGUMENT;
  
  }
  
  if(imap->procs!=NULL){
     *procs=realloc(imap->procs,imap->ttable_ic_size*sizeof(*(imap->procs)));
  }
  else{
   *procs=NULL;
  }
  
  if(imap->local_adr!=NULL){
    *local_adr=realloc(imap->local_adr,imap->ttable_ic_size*sizeof(*(imap->local_adr)));
  }
  else{
    *local_adr=NULL;
  }
  
  *ttable_size=imap->ttable_ic_size;
  
  imap->ttable_ic_size=0;
  
  imap->ttable_ic_real_size=0;
  
  imap->procs=NULL;
  
  imap->local_adr=NULL;
  return LIP_SUCCESS;
}

int LIP_Datamap_ic_ttable_free(LIP_Datamap datamap){
  LIP_IDatamap *imap;
  
  int *iprocs;
  
  int *ilocal_adr;
  
  char* pname="LIP_Datamap_ic_ttable_free";  
  
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can free translation table only of INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if(NULL!=imap->procs){
   free(imap->procs);
   imap->procs=NULL;
  }
  if(NULL!=imap->local_adr){
   free(imap->local_adr);
   imap->local_adr=NULL;
  }
  imap->ttable_ic_size=0;
  
  imap->ttable_ic_real_size=0;
  
  return LIP_SUCCESS;
}
/*zmiana*/
/*
flags
LIP_COPY -processors table is copied - requires more time and memory,
but one can change proc array afterwords
LIP_ATTACH- a pointer to the processors table is copied-
one should not change procs array afterwords 
*/

int LIP_Datamap_ic_init_table_with_proc(LIP_Datamap datamap,void*procs,
  LIP_Indextype procstype,
  int ttable_size, int flags ){
  LIP_IDatamap *imap;
  
  int *iprocs,i;
  
  char* pname="LIP_Datamap_ic_init_ttable_with_proc";  
  
  if (procstype != LIP_INDEXTYPE_INT && NULL!=procs
     ){
    LIP_Error_ret(pname,"indextype",LIP_NOT_SUPPORTED);
    return LIP_NOT_SUPPORTED;
  }   
  if(procs==NULL){
    LIP_Error_ret(pname,"cannot fill translation table with null processors' array",LIP_INVALID_ARGUMENT);  
    return LIP_INVALID_ARGUMENT;
  } 
  iprocs = procs;
  
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can fill procs table only to INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  
  imap->ttable_ic_size=ttable_size;
  if(imap->flag==LIP_NONLOCAL){
      imap->ttable_ic_real_size=ttable_size*GHOST_AREA_RATE;
  }
  else{
      imap->ttable_ic_real_size=ttable_size;
  }
  
  if(flags==LIP_COPY){
    /*processors table is copied - requires more time and memory,
    but one can change proc array afterwords*/

    if(imap->procs==NULL){
     imap->procs=malloc(imap->ttable_ic_real_size*sizeof(*iprocs));
    } 
    else{
     imap->procs=realloc(imap->procs,imap->ttable_ic_real_size*sizeof(*iprocs));
    
    } 

    memcpy( imap->procs, procs, imap->ttable_ic_size * (sizeof *iprocs) );
  }
  if(LIP_ATTACH==flags){
   /*a pointer to the processors table is copied
   one should not change procs array afterwords */
   
   imap->procs=realloc(procs,imap->ttable_ic_real_size*sizeof(*iprocs));
  
  
  }
 return LIP_SUCCESS;
}

/*ZMIANA*/
/*
flags
LIP_COPY - arraysof translation table are copied - requires more time and memory,
but one can change proc array afterwords
LIP_ATTACH- a pointer to arrays of translation table are copied-
one should not change procs array afterwords 
*/

int LIP_Datamap_ic_ttable_add(LIP_Datamap datamap,
  void *procs, LIP_Indextype procstype,
  void *local_adr,LIP_Indextype localtype,
  int ttable_size,int flag){

  LIP_IDatamap *imap;
  
  int *iprocs;
  int *ilocal_adr;
  
  char* pname="LIP_Datamap_ic_ttable_attach";  
  if (( procstype != LIP_INDEXTYPE_INT && NULL!=procs) ||
      (localtype != LIP_INDEXTYPE_INT && NULL!=local_adr)
     ){
    LIP_Error_ret(pname,"indextype",LIP_NOT_SUPPORTED);
    return LIP_NOT_SUPPORTED;
  }    
  iprocs = procs;
  ilocal_adr = local_adr;
  
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can attach table only to INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  
  imap->ttable_ic_size=ttable_size;
  if(LIP_ATTACH==flag){
      if(LIP_NONLOCAL==imap->flag){
  
       imap->ttable_ic_real_size=GHOST_AREA_RATE*ttable_size;
  
       if(NULL!=procs)
           imap->procs=realloc(iprocs,(imap->ttable_ic_real_size)* sizeof(*iprocs));
       if(NULL==procs){
	    LIP_Error_ret(pname,"cannot allocate memory for ghost area in translation table - try to decrease GHOST_AREA_RATE",LIP_OUT_OF_MEMORY);
	     return LIP_OUT_OF_MEMORY;
       }
       if(NULL!=local_adr)
           imap->local_adr=realloc(ilocal_adr,(imap->ttable_ic_real_size)* sizeof(*ilocal_adr));
       if (NULL==local_adr){
	    LIP_Error_ret(pname,"cannot allocate memory for ghost area in translation table - try to decrease GHOST_AREA_RATE",LIP_OUT_OF_MEMORY);
	    return LIP_OUT_OF_MEMORY;
       }
      }
      if(LIP_LOCAL==imap->flag){
  
       imap->ttable_ic_real_size=ttable_size;
  
       imap->procs=procs;
  
       imap->local_adr=local_adr;
  
      }
  }
  else if (LIP_COPY==flag){
      if(LIP_NONLOCAL==imap->flag){
         imap->ttable_ic_size=ttable_size;
         imap->ttable_ic_real_size=GHOST_AREA_RATE*ttable_size;
      }
      else if(LIP_LOCAL==imap->flag){
	 imap->ttable_ic_size=ttable_size;
         imap->ttable_ic_real_size=ttable_size;
      }
      if(procs!=NULL){
          if(imap->procs==NULL){
              imap->procs=malloc(imap->ttable_ic_real_size*sizeof(*iprocs));
          }
	  else{
	      imap->procs=realloc(imap->procs,imap->ttable_ic_real_size*sizeof(*iprocs));
          }
	  memcpy(imap->procs,procs,imap->ttable_ic_size*sizeof(*iprocs));
      }
      if(local_adr!=NULL){
          if(imap->local_adr==NULL){
              imap->local_adr=malloc(imap->ttable_ic_real_size*sizeof(*ilocal_adr));
          }
	  else{
	      imap->local_adr=realloc(imap->local_adr,imap->ttable_ic_real_size*sizeof(*ilocal_adr));
          }
	  memcpy(imap->local_adr,local_adr,imap->ttable_ic_size*sizeof(*ilocal_adr));
      }
  }
  
  return LIP_SUCCESS;
}

int LIP_Datamap_ooc_ttable_attach(
  LIP_Datamap datamap,
  char * procs_fname,
  char * local_adr_fname,
  int ttable_size,
  int ttable_membuf_size){
  int ret;
  LIP_IDatamap *imap;
  char*pname="LIP_Datamap_ooc_ttable_attach";  
 
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can attach table only to INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  
  imap->ttable_ooc_size=ttable_size;
  imap->procs_fname= malloc (strlen (procs_fname));
  imap->local_adr_fname= malloc (strlen (local_adr_fname));
  strcpy(local_adr_fname, imap->local_adr_fname);
  strcpy(procs_fname, imap->procs_fname);
  MPI_File_open(LIP_env.comm, procs_fname, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &imap->procs_file);
  MPI_File_open(LIP_env.comm, local_adr_fname, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &imap->local_adr_file);

  imap->ttable_ic_size=ttable_membuf_size;
  if (LIP_NONLOCAL==imap->flag){
   imap->ttable_ic_real_size=GHOST_AREA_RATE*ttable_membuf_size;
  }
  else{
   imap->ttable_ic_real_size=ttable_membuf_size;
  }
  
  imap->procs=malloc(imap->ttable_ic_real_size*sizeof(int));

  if(NULL==imap->procs){
    LIP_Error_ret(pname,"cannot allocate memory for translation table page",LIP_OUT_OF_MEMORY);
    return LIP_OUT_OF_MEMORY;
  }
  imap->local_adr=malloc(imap->ttable_ic_real_size*sizeof(int));
  
  if(NULL==imap->local_adr){
    LIP_Error_ret(pname,"cannot allocate memory for translation table page",LIP_OUT_OF_MEMORY);
    return LIP_OUT_OF_MEMORY;
  }  
  
  ret=LIP_IOBufmap_create(imap->ttable_ic_size,&(imap->tt_io_bmap));
  
  if(LIP_SUCCESS!=ret){
    LIP_Error_ret(pname,NULL,ret);
    return ret;
  }
  return LIP_SUCCESS;

}

/* This function gets names of files containing
   ooc translatin table from datamap object.
   procs_fname and local_adr_fname arguments
   are set to newly allocated strings.
*/

int LIP_Datamap_ooc_to_local(
  LIP_Datamap datamap,
  char* local_dir)

{
  char* new_proc_name,*local_dir1,*local_dir2;
  char* new_local_name;		
  LIP_IDatamap *imap;
  char*pname="LIP_Datamap_ooc_to_local";  
  MPI_File local_adr_file, procs_file;
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can attach table only to INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }


local_dir1=malloc(sizeof(char)*(strlen(local_dir)+strlen(imap->procs_fname)+1));
local_dir2=malloc(sizeof(char)*(strlen(local_dir)+strlen(imap->local_adr_fname)+1));
  local_dir1=strcpy(local_dir1,local_dir);
  local_dir2=strcpy(local_dir2,local_dir);
  new_proc_name=strcat(local_dir1,(imap->procs_fname)+1);
  new_local_name=strcat(local_dir2,(imap->local_adr_fname)+1);


MPI_File_open(MPI_COMM_SELF, new_proc_name, MPI_MODE_RDWR|
                          MPI_MODE_CREATE, MPI_INFO_NULL,&procs_file);

MPI_File_open(MPI_COMM_SELF,new_local_name,MPI_MODE_RDWR |
                          MPI_MODE_CREATE, MPI_INFO_NULL,
                &local_adr_file);



LIP_copy_files(imap->procs_file,procs_file,imap->ttable_ooc_size*LIP_env.rank,
               imap->ttable_ooc_size,imap->ttable_ic_size,MPI_INT);

LIP_copy_files(imap->local_adr_file,local_adr_file,imap->ttable_ooc_size*LIP_env.rank,
               imap->ttable_ooc_size,imap->ttable_ic_size,MPI_INT);

MPI_File_close(&imap->procs_file);

MPI_File_delete(imap->procs_fname,MPI_INFO_NULL);

imap->procs_fname=new_proc_name;
imap->procs_file=procs_file;

MPI_File_close(&imap->local_adr_file);

MPI_File_delete(imap->local_adr_fname,MPI_INFO_NULL);

imap->procs_fname=new_local_name;
imap->local_adr_file=local_adr_file;

return LIP_SUCCESS;
}
int LIP_Datamap_ooc_ttable_get_files(
  LIP_Datamap datamap,
  char * procs_fname,
  char * local_adr_fname)
  {
  int ret;
  LIP_IDatamap *imap;
  char*pname="LIP_Datamap_ooc_get_files";  
 
  imap=LIP_Datamap_convert(datamap);
  if (imap==LIP_DATAMAP_NULL){
   LIP_Error_ret(pname,"Datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (imap->type!=LIP_DATAMAP_INDIRECT){
   LIP_Error_ret(pname,"you can get files only from INDIRECT datamap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  if (NULL==imap->procs_fname||NULL==imap->local_adr_fname){
   LIP_Error_ret(pname,"no files attached!",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  
  procs_fname= malloc (strlen (imap->procs_fname));
  local_adr_fname= malloc (strlen (imap->local_adr_fname));
  strcpy(imap->local_adr_fname, local_adr_fname);
  strcpy(imap->procs_fname, procs_fname);

  return LIP_SUCCESS;

}




static LIP_IDatamap *datamap_array[MAX_DATAMAP];
static int datamap_index = 0;
int LIP_Datamap_array_init(){
int i;
for (i=0;i<MAX_DATAMAP;i++){
 datamap_array[i]=NULL;
}
return LIP_SUCCESS;
}
int LIP_Datamap_array_free(){
int i;
LIP_Datamap dm;
for (i=0;i<MAX_DATAMAP;i++){
 if(datamap_array[i]!=NULL){
  dm=i+LIP_DATAMAP_FIRST;
  LIP_Datamap_free(&dm, 0);
 }
}

return LIP_SUCCESS;
}
LIP_IDatamap *
LIP_Datamap_new(LIP_Datamap *datamap)
{
  LIP_IDatamap *map;
  int found,i;
  char*pname="LIP_Datamap_new";
  
  *datamap = LIP_DATAMAP_NULL;
  if(datamap_array[datamap_index]!=NULL){
   found=0;
   for(i=0;i<MAX_DATAMAP;i++){
    if (NULL==datamap_array[i]){
      datamap_index=i;
      found=1;
      break;
    }
   }
   if(!found){
     LIP_Error_ret(pname,"MAX_DATAMAP too small",LIP_TOO_MANY_HANDLES);
     return NULL;
   }
  }

  map = datamap_array[datamap_index] = malloc( sizeof *map );
  
  if (NULL == map){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return NULL;
  }
  
  *datamap = datamap_index++ + LIP_DATAMAP_FIRST;
  
  datamap_index=datamap_index%MAX_DATAMAP;
  
  return map;
}


int
LIP_Datamap_free(LIP_Datamap *dmap, int flags)
{
MPI_Info info;
LIP_IDatamap *idm;
  char* pname="LIP_Datamap_free";
  if (LIP_DATAMAP_NULL == *dmap){
    LIP_Error_ret(pname,"trying to free NULL Datamap",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
  idm=LIP_Datamap_convert(*dmap);

  if(NULL==idm){
    LIP_Error_ret(pname,"trying to free invalid Datamap",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }  

  if(NULL!=idm->procs){
   free(idm->procs);
   idm->procs=NULL;
  }

  if(NULL!=idm->local_adr){
   free(idm->local_adr);
   idm->local_adr=NULL;
  
  }
  
  if(LIP_IOBUFMAP_NULL!=idm->tt_io_bmap){
     LIP_IOBufmap_free(&(idm->tt_io_bmap));
     idm->tt_io_bmap=LIP_IOBUFMAP_NULL; 
  }
  
  if (MPI_FILE_NULL!=idm->procs_file){
        MPI_File_close(&idm->procs_file);
        if (! (flags & LIP_DATAMAP_STORE)) 
             MPI_File_delete(idm->procs_fname, info);
  }

  if(MPI_FILE_NULL!=idm->local_adr_file){
        MPI_File_close(&idm->local_adr_file);
        if (! (flags & LIP_DATAMAP_STORE))
             MPI_File_delete(idm->local_adr_fname, info);
  }
  
  datamap_array[(*dmap)-LIP_DATAMAP_FIRST]=NULL;
  
  *dmap = LIP_DATAMAP_NULL;

  return LIP_SUCCESS;
}

LIP_IDatamap *
LIP_Datamap_convert(LIP_Datamap m)
{
  return LIP_DATAMAP_NULL == m ? NULL :
    m < LIP_DATAMAP_FIRST ? NULL : datamap_array[m - LIP_DATAMAP_FIRST];
}

/* 
 Side effects: collective communication when mapping is irregular
 */
int
LIP_Dereference(LIP_Datamap datamap, void *iglobal, LIP_Indextype igtype,
  void *ilocal, LIP_Indextype iltype, int *procs, int count, int deref_flag,
  int* indcs_on_proc)
{
  int i,pot_new_real_size,ret;
  int *iglobal_int, *ilocal_int,*itocal;
  int *tt_local_ghost,*tt_procs_ghost;
  LIP_IDatamap *imap;
  LIP_Datamap tt_dmap;
  LIP_Schedule tt_sch;
  LIP_ISchedule *tt_isch;
  MPI_Status status;
  MPI_Datatype memtype,filetype;
  MPI_Info info; 
  int quot, quot1, rem; /* quotient, quotient+1, reminder */
  
  char*pname="LIP_Dereference";
  int fl_indx; /* first linear index */
  MPI_Info_create(&info);
  /* FIXME: check arguments */
  /* FIXME: general indexing type should be supported */
  if (LIP_INDEXTYPE_INT != igtype || LIP_INDEXTYPE_INT != iltype){
   LIP_Error_quit(pname,"indextype",LIP_NOT_SUPPORTED);
  }
  else
  {
    iglobal_int = iglobal;
    ilocal_int  = ilocal;
  }

  if (1 == LIP_env.size)
  {
    memcpy( ilocal_int, iglobal_int, count * (sizeof *ilocal_int) );

    memset( procs, 0, count * (sizeof *procs) );
    /* this is equivalent to: */
    /*
    for (i = indcs_size - 1 ; i >= 0 ; i--)
      procs[i] = 0;
    */

    indcs_on_proc[0] = count;

    return LIP_SUCCESS;
  }

  /* FIXME: check flags if counters should be filled */

  /* iterate over all counters of local indices (for each node) */
  for (i = LIP_env.size - 1 ; i >= 0 ; i--)
    indcs_on_proc[i] = 0;

  switch (datamap)
  {
    case LIP_CYCLIC:
    /* FIXME: use a dynamic data map for CYCLIC(1) */
    break;

    default:
    imap = LIP_Datamap_convert( datamap );
    break;
  }

  switch (imap->type)
  {
    /* HPF 2.0 BLOCK(m) */
    case LIP_DATAMAP_BLOCK:
      for (i = count - 1 ; i >= 0 ; i--)
      {
        procs[i] = CEILDIV( iglobal_int[i], imap->local_size );
        ilocal_int[i] = CEILREM( iglobal_int[i], imap->local_size );
        indcs_on_proc[procs[i]]++;
      }
    break;
    /* IRREGULAR MAPPING*/
    case LIP_DATAMAP_INDIRECT:
      /*IN-CORE TRANSLATION TABLE*/
      if(0==imap->ttable_ooc_size){
       
       ret=LIP_Schedule_create(imap->ttable_ic_size,&tt_sch);
       
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }
       
       ret=LIP_Datamap_create(LIP_DATAMAP_BLOCK,imap->global_size,&imap->local_size,
          LIP_INDEXTYPE_INT,1,0,&tt_dmap);
       
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }
       
       ret=LIP_Localize(tt_dmap,iglobal,igtype,ilocal,iltype,
       count,0,&tt_sch,LIP_INDEXINFO_NULL);
       
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }
       
       ret=LIP_Schedule_commit(&tt_sch);
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }
       
       LIP_Gather(imap->procs,imap->procs+imap->ttable_ic_size,MPI_INT,tt_sch);       
       
       LIP_Gather(imap->local_adr,imap->local_adr+imap->ttable_ic_size,MPI_INT,tt_sch);       
       
       for (i = count - 1 ; i >= 0 ; i--)
       {
         procs[i] = imap->procs[ilocal_int[i]];
         ilocal_int[i] = imap->local_adr[ilocal_int[i]];
       	 indcs_on_proc[procs[i]]++;
       }
       
       LIP_Schedule_free(&tt_sch);
       LIP_Datamap_free(&tt_dmap, 0);
       
     }
     /*OOC TRANSLATION TABLE*/
     else{

       ret=LIP_Schedule_create(imap->ttable_ooc_size,&tt_sch);
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       ret=LIP_Datamap_create(LIP_DATAMAP_BLOCK,imap->global_size,
                          &imap->local_size,
                          LIP_INDEXTYPE_INT,1,0,
			   &tt_dmap);
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       ret=LIP_Localize(tt_dmap,iglobal,igtype,ilocal,iltype,
       count,0,&tt_sch,LIP_INDEXINFO_NULL);

       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       ret=LIP_OOC_localize(ilocal,iltype,ilocal,iltype,count,0,NULL,
                        MPI_INT,&(imap->tt_io_bmap),&tt_sch);
       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       ret= LIP_Schedule_commit(&tt_sch);

       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       ret=LIP_IOBufmap_get_datatype(imap->tt_io_bmap,MPI_INT,&memtype,&filetype,
                                 LIP_DATA_NEW);

       if(LIP_SUCCESS!=ret){
        LIP_Error_quit(pname,NULL,ret);
       }

       if(MPI_DATATYPE_NULL!=memtype){
        if(LIP_LOCAL==imap->flag){

        	MPI_File_set_view(imap->procs_file,
			 (imap->ttable_ooc_size)*LIP_env.rank*sizeof(int),
			  MPI_INT,filetype,"native",info);

        	MPI_File_set_view(imap->local_adr_file,
	                  (imap->ttable_ooc_size)*LIP_env.rank*sizeof(int),
			  MPI_INT,filetype,"native",info);

	 	MPI_File_read_at(imap->local_adr_file,0,imap->local_adr,1,memtype,
	                 &status);		  

	  	MPI_File_read_at(imap->procs_file,0,imap->procs,1,memtype,
	                 &status);		  
        }
        else{
		MPI_File_set_view(imap->procs_file,0,
			  MPI_INT,filetype,"native",info);

        	MPI_File_set_view(imap->local_adr_file,0,
			  MPI_INT,filetype,"native",info);

	 	MPI_File_read_at(imap->local_adr_file,0,imap->local_adr,1,memtype,
	                 &status);		  

	  	MPI_File_read_at(imap->procs_file,0,imap->procs,1,memtype,
	                 &status);		  
      	
	}
        MPI_Type_free(&memtype);
        MPI_Type_free(&filetype);
       }
       else{
        MPI_File_set_view(imap->procs_file,
				0,
			  MPI_INT,MPI_INT,"native",info);

        MPI_File_set_view(imap->local_adr_file,0,
			  MPI_INT,MPI_INT,"native",info);
    
	

	MPI_File_read_at(imap->local_adr_file,0,imap->local_adr,0,MPI_INT,
	                 &status);		  

	MPI_File_read_at(imap->procs_file,0,imap->procs,0,MPI_INT,
	                 &status);		  
       
       }				
       LIP_Gather(imap->procs,imap->procs+imap->ttable_ic_size,MPI_INT,tt_sch);       
       LIP_Gather(imap->local_adr,imap->local_adr+imap->ttable_ic_size,MPI_INT,tt_sch);       
       for (i = count - 1 ; i >= 0 ; i--)
       {
#ifdef DEBUG
fprintf(debf,"%d\n",i);
#endif
         procs[i] = imap->procs[ilocal_int[i]];
         ilocal_int[i] = imap->local_adr[ilocal_int[i]];
         indcs_on_proc[procs[i]]++;
       }
       LIP_Schedule_free(&tt_sch);
#ifdef DEBUG
fprintf(debf,"pofreedch\n");
#endif

       LIP_Datamap_free(&tt_dmap, 0);
#ifdef DEBUG
fprintf(debf,"pofreedatamap\n");
#endif
     }
     break;
    /* FIXME: evenly blocked not in HPF 2.0 */
    default:
      /* each node has at least 'quot' data elements */
      quot = imap->global_size / LIP_env.size;

      /* there might be some extra indices; 'rem' is at the same time first
       * linear node, i.e. the node from which nodes have 'quot' data elements
       * (without one extra index) */
      rem   = imap->global_size % LIP_env.size;

      /* first 'rem' nodes will have 'quot1' data elements */
      quot1 = quot + 1;

      /* first linear index is the first index of data element on the first
       * node that has 'quot' indices, all of the previous nodes have 'quot1'
       * data elements */
      fl_indx = rem * quot + rem;

      /* for all indices */
      for (i = count - 1 ; i >= 0 ; i--)
      {
        if (iglobal_int[i] < fl_indx)
        {
          procs[i]   = iglobal_int[i] / quot1;
          ilocal_int[i] = iglobal_int[i] % quot1;
        }
        else
        {
          procs[i]   = (iglobal_int[i] - fl_indx) / quot + rem;
          ilocal_int[i] = (iglobal_int[i] - fl_indx) % quot;
        }

        indcs_on_proc[procs[i]]++;
      }
    break;
  }

  return LIP_SUCCESS;
} /* LIP_Dereference */

